/****************************************************************
 * Copyright (c) 2003-2005, Cavium Networks. All rights reserved.
 *
 * This Software is the property of Cavium Networks.  The Software and all
 * accompanying documentation are copyrighted.  The Software made available
 * here constitutes the proprietary information of Cavium Networks.  You
 * agree to take reasonable steps to prevent the disclosure, unauthorized use
 * or unauthorized distribution of the Software.  You shall use this Software
 * solely with Cavium hardware.
 *
 * Except as expressly permitted in a separate Software License Agreement
 * between You and Cavium Networks, you shall not modify, decompile,
 * disassemble, extract, or otherwise reverse engineer this Software.  You
 * shall not make any copy of the Software or its accompanying documentation,
 * except for copying incident to the ordinary and intended use of the
 * Software and the Underlying Program and except for the making of a single
 * archival copy.
 *
 * This Software, including technical data, may be subject to U.S.  export
 * control laws, including the U.S.  Export Administration Act and its
 * associated regulations, and may be subject to export or import regulations
 * in other countries.  You warrant that You will comply strictly in all
 * respects with all such regulations and acknowledge that you have the
 * responsibility to obtain licenses to export, re-export or import the
 * Software.
 *
 * TO THE MAXIMUM EXTENT PERMITTED BY LAW, THE SOFTWARE IS PROVIDED "AS IS"
 * AND WITH ALL FAULTS AND CAVIUM MAKES NO PROMISES, REPRESENTATIONS OR
 * WARRANTIES, EITHER EXPRESS, IMPLIED, STATUTORY, OR OTHERWISE, WITH RESPECT
 * TO THE SOFTWARE, INCLUDING ITS CONDITION, ITS CONFORMITY TO ANY
 * REPRESENTATION OR DESCRIPTION, OR THE EXISTENCE OF ANY LATENT OR PATENT
 * DEFECTS, AND CAVIUM SPECIFICALLY DISCLAIMS ALL IMPLIED (IF ANY) WARRANTIES
 * OF TITLE, MERCHANTABILITY, NONINFRINGEMENT, FITNESS FOR A PARTICULAR
 * PURPOSE, LACK OF VIRUSES, ACCURACY OR COMPLETENESS, QUIET ENJOYMENT, QUIET
 * POSSESSION OR CORRESPONDENCE TO DESCRIPTION.  THE ENTIRE RISK ARISING OUT
 * OF USE OR PERFORMANCE OF THE SOFTWARE LIES WITH YOU.
 *
 **************************************************************************/

/**
 * @file oct-linux-csr.c
 *
 * Utility to display CSR under Linux
 *
 * $Id: oct-linux-csr.c,v 1.2 2006/07/25 02:13:18 kreese Exp $ $Name: SDK_1_5_0_build_187 $
 *
 *
 */
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include "cvmx.h"
typedef uint64_t uint64;
#include "octeon-pci-csr-db.h"


/**
 * Get the address for the named CSR
 *
 * @param name   Name of the CSR to get
 * @return Address or zero on success
 */
static const OCTEON_CSR_ADDRESS_TYPE *cvmx_csr_get(const char *name)
{
    int i;
    for (i=0; i<OCTEON_CSR_ADDRESS_NUM; i++)
    {
        if (strcasecmp(name, octeon_csr_addresses[i].name) == 0)
            return octeon_csr_addresses + i;
    }
    return NULL;
}


/**
 * Print a decoded CSR
 *
 * @param address Address of CSR to decode
 * @param value   Value to decode
 */
static void cvmx_csr_decode(uint64_t address, uint64_t value)
{
    int i;
    for (i=0; i<OCTEON_CSR_ADDRESS_NUM; i++)
    {
        if (octeon_csr_addresses[i].address == address)
        {
            int csr = octeon_csr_addresses[i].csroff;
            printf("%s = 0x%016llx\n", octeon_csr_addresses[i].name, (unsigned long long)value);
            for (i=octeon_csrs[csr].fieldoff+octeon_csrs[csr].numfields-1; i>=octeon_csrs[csr].fieldoff; i--)
            {
                uint64_t v = (value >> octeon_csr_fields[i].startbit);
                v = v & ~((~0x0ull) << octeon_csr_fields[i].sizebits);
                printf("    %-20s = %10llu (0x%llx)\n",
                       octeon_csr_fields[i].name,
                       (unsigned long long)v,
                       (unsigned long long)v);
            }
            break;
        }
    }
}


/**
 * Print a list of all matching CSR names
 *
 * @param prefix All CSR starting with this are printed
 */
static void cvmx_csr_display_list(const char *prefix)
{
    int i, len;
    if (prefix == NULL)
        prefix = "";
    len = strlen(prefix);

    for (i=0; i<OCTEON_CSR_ADDRESS_NUM; i++)
    {
        if (strncasecmp(prefix, octeon_csr_addresses[i].name, len) == 0)
            printf("%s\n", octeon_csr_addresses[i].name);
    }
}

uint64_t convert_number(const char *str)
{
    unsigned long long result;
    if ((str[0] == '0') && (str[1] == 'x'))
    {
        if (sscanf(str+2, "%llx", &result) != 1)
            result = 0;
    }
    else
    {
        if (sscanf(str, "%lli", &result) != 1)
            result = 0;
    }
    return result;
}

/**
 * Main
 *
 * @param argc
 * @param argv
 * @return
 */
int main(int argc, const char *argv[])
{
    const char *csr_name;

    /* Make sure we got the correct number of arguments */
    if ((argc != 2) && (argc != 3))
    {
        printf("\n"
               "Usage: oct-linux-csr <CSR> [value]\n"
               "    Read or write a CSR by name.\n"
               "\n"
               "    CSR     Name of CSR to read or write.\n"
               "    value   If present, write this value to the CSR. Otherwise do a read.\n"
               "\n"
               "    If CSR is a partial name, the list of CSRs starting with it are displayed.\n"
               "\n");
        return -1;
    }

    /* First argument is the CSR we're interested in */
    csr_name = argv[1];

    if (!strncasecmp(csr_name, "CVMX_", 5))
    {
        csr_name += 5;
    }
    /* Find the CSR address from the name. If this fails it will return 0 */
    const OCTEON_CSR_ADDRESS_TYPE *csr = cvmx_csr_get(csr_name);

    if (csr)
    {
        /* Determine if we're doing a read or a write */
        if (argc == 3)
        {
            if (csr->widthbits == 32)
                cvmx_write64_uint32((csr->address^4)|(1ull<<63), convert_number(argv[2]));
            else
                cvmx_write64_uint64(csr->address|(1ull<<63), convert_number(argv[2]));
        }
        else
        {
            uint64_t data;
            if (csr->widthbits == 32)
                data = cvmx_read64_uint32((csr->address^4)|(1ull<<63));
            else
                data = cvmx_read64_uint64(csr->address|(1ull<<63));

            cvmx_csr_decode(csr->address, data);
        }
    }
    else
    {
        printf("CSR of name: %s not found.\n", csr_name);
        cvmx_csr_display_list(csr_name);
    }
    return 0;
}

