/****************************************************************
 * Copyright (c) 2005-2006, Cavium Networks. All rights reserved.
 *
 * This Software is the property of Cavium Networks.  The Software and all
 * accompanying documentation are copyrighted.  The Software made available
 * here constitutes the proprietary information of Cavium Networks.  You
 * agree to take reasonable steps to prevent the disclosure, unauthorized use
 * or unauthorized distribution of the Software.  You shall use this Software
 * solely with Cavium hardware.
 *
 * Except as expressly permitted in a separate Software License Agreement
 * between You and Cavium Networks, you shall not modify, decompile,
 * disassemble, extract, or otherwise reverse engineer this Software.  You
 * shall not make any copy of the Software or its accompanying documentation,
 * except for copying incident to the ordinary and intended use of the
 * Software and the Underlying Program and except for the making of a single
 * archival copy.
 *
 * This Software, including technical data, may be subject to U.S.  export
 * control laws, including the U.S.  Export Administration Act and its
 * associated regulations, and may be subject to export or import regulations
 * in other countries.  You warrant that You will comply strictly in all
 * respects with all such regulations and acknowledge that you have the
 * responsibility to obtain licenses to export, re-export or import the
 * Software.
 *
 * TO THE MAXIMUM EXTENT PERMITTED BY LAW, THE SOFTWARE IS PROVIDED "AS IS"
 * AND WITH ALL FAULTS AND CAVIUM MAKES NO PROMISES, REPRESENTATIONS OR
 * WARRANTIES, EITHER EXPRESS, IMPLIED, STATUTORY, OR OTHERWISE, WITH RESPECT
 * TO THE SOFTWARE, INCLUDING ITS CONDITION, ITS CONFORMITY TO ANY
 * REPRESENTATION OR DESCRIPTION, OR THE EXISTENCE OF ANY LATENT OR PATENT
 * DEFECTS, AND CAVIUM SPECIFICALLY DISCLAIMS ALL IMPLIED (IF ANY) WARRANTIES
 * OF TITLE, MERCHANTABILITY, NONINFRINGEMENT, FITNESS FOR A PARTICULAR
 * PURPOSE, LACK OF VIRUSES, ACCURACY OR COMPLETENESS, QUIET ENJOYMENT, QUIET
 * POSSESSION OR CORRESPONDENCE TO DESCRIPTION.  THE ENTIRE RISK ARISING OUT
 * OF USE OR PERFORMANCE OF THE SOFTWARE LIES WITH YOU.
 *
 **************************************************************************/

/**
 * @file oct-pci-profile.c
 *
 * Octeon profiler under Linux (uses up one core)
 *
 * $Id: oct-linux-profile.cpp,v 1.1 2006/10/11 16:44:15 kreese Exp $ $Name: SDK_1_5_0_build_187 $
 */
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <getopt.h>
#include <map>
#include "cvmx.h"
#include "../../bootloader/u-boot/include/octeon_mem_map.h"
#include "octeon-pci-debug.h"

typedef struct
{
    uint32_t count[16];
} ProfileElement;

typedef std::map<uint64_t,ProfileElement> ProfileData;

static ProfileData      profile_data;
static int              fast = 0;
static int              dwell = 5;

int parse_options(int argc, char * const *argv)
{
    const struct option long_options[] =
    {
        {"fast", 0, 0, 'e'},
        {"dwell", 1, 0, 'f'},
        {NULL, 0, 0, 0}
    };

    int option_index = 0;
    int show_usage = 0;

    while (1)
    {
        char option = getopt_long(argc, argv, "ef:", long_options, &option_index);
        if (option <= 0)
            break;

        switch (option)
        {
            case 'e':
                fast = 1;
                break;

            case 'f':
                dwell = atoi(optarg);
                break;

            default:
                show_usage = 1;
                break;
        }
    }

    if (show_usage)
    {
        printf("\n"
               "Cavium Octeon Linux Profiler\n"
               "\n"
               "Usage:\n"
               "    %s [options]\n"
               "\n"
               "Supported Options:\n"
               "    --fast\n"
               "    --dwell=<seconds>\n", argv[0]);
        printf("\n\n");
        return -1;
    }
    else
        return 0;
}

int main(int argc, char * const *argv)
{
    int core;
    int number_of_cores;
    int coremask;

    if (parse_options(argc, argv))
        return -1;

    uint64_t fuse = cvmx_read_csr(CVMX_CIU_FUSE);
    number_of_cores = 0;
    while (fuse&1)
    {
        number_of_cores++;
        fuse>>=1;
    }
    coremask = (1<<number_of_cores)-1;

    uint32_t handler = *(volatile uint32_t *)CVMX_ADD_IO_SEG(BOOTLOADER_DEBUG_HANDLER_LIST+OCTEON_PCI_DEBUG_HANDLER_PCI_PROFILER*4);
    *(volatile uint32_t *)CVMX_ADD_IO_SEG(BOOTLOADER_DEBUG_TRAMPOLINE) = handler;
    CVMX_SYNCW;
    usleep(100000);

    time_t start_time = time(NULL);
    while (1)
    {
        cvmx_write_csr(CVMX_CIU_DINT, coremask);

        /* Put delay after core stop so all core writes should be done before
            we read. Since the cores automatically startup again, this delay
            doesn't affect Octeon */
        if (!fast)
            usleep(1000);

        for (core=0; core<number_of_cores; core++)
        {
            uint64_t address = CVMX_ADD_IO_SEG(BOOTLOADER_DEBUG_REG_SAVE_BASE + (core<<12) + offsetof(octeon_pci_debug_t, cop0_depc));
            uint64_t pc = *(volatile uint64_t *)address;
            profile_data[pc].count[core]++;
        }

        if (time(NULL) - start_time >= dwell)
        {
            ProfileData::const_iterator iter;
            for (iter=profile_data.begin(); iter!=profile_data.end(); iter++)
            {
                printf("0x%016llx:", (unsigned long long)iter->first);
                for (core=0; core<number_of_cores; core++)
                    printf(" %5u", iter->second.count[core]);
                printf("\n");
            }
            printf("\n");
            fflush(stdout);
            profile_data.clear();
            start_time = time(NULL);
        }
    }

    return 0;
}

