/********************************************************************\
 * This program is free software; you can redistribute it and/or    *
 * modify it under the terms of the GNU General Public License as   *
 * published by the Free Software Foundation; either version 2 of   *
 * the License, or (at your option) any later version.              *
 *                                                                  *
 * This program is distributed in the hope that it will be useful,  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU General Public License for more details.                     *
 *                                                                  *
 * You should have received a copy of the GNU General Public License*
 * along with this program; if not, contact:                        *
 *                                                                  *
 * Free Software Foundation           Voice:  +1-617-542-5942       *
 * 59 Temple Place - Suite 330        Fax:    +1-617-542-2652       *
 * Boston, MA  02111-1307,  USA       gnu@gnu.org                   *
 *                                                                  *
 \********************************************************************/

/* $Header: /cvsroot/wifidog/wifidog/src/http.c,v 1.37 2005/03/20 21:22:24 minaguib Exp $ */
/** @file http.c
  @brief HTTP IO functions
  @author Copyright (C) 2004 Philippe April <papril777@yahoo.com>
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <string.h>
#include <unistd.h>
#include <syslog.h>

#include "httpd.h"

#include "safe.h"
#include "debug.h"
#include "conf.h"
#include "auth.h"
#include "firewall.h"
#include "http.h"
#include "httpd.h"
#include "client_list.h"
#include "common.h"

#include "util.h"

#include "../config.h"

#define IS_NULL(x)  ((x) ? (x) : "<NULL>")

extern pthread_mutex_t	client_list_mutex;

typedef struct __t_redir_node {
	struct __t_redir_node *next;
	char *mac;
	char redir_pending;
	time_t expiry;
} t_redir_node;

t_redir_node *first_redir_node = NULL;
int redir_pending = 0;
pthread_mutex_t redir_mutex = PTHREAD_MUTEX_INITIALIZER;

#define LOCK_REDIR() do { \
	        pthread_mutex_lock(&redir_mutex); \
} while (0)

#define UNLOCK_REDIR() do { \
	        pthread_mutex_unlock(&redir_mutex); \
} while (0)

t_redir_node *
redir_list_find(char *mac)
{
	t_redir_node *ptr;

	ptr = first_redir_node;
	while (NULL != ptr) {
		if (!strcmp(ptr->mac, mac))
			return ptr;
		ptr = ptr->next;
	}

	return NULL;
}


t_redir_node *redir_list_append(char *mac)
{
	t_redir_node *curnode, *prevnode;

	prevnode = NULL;
	curnode = first_redir_node;

	while (curnode != NULL) {
		prevnode = curnode;
		curnode = curnode->next;
	}

	curnode = malloc(sizeof(t_redir_node));
	if (curnode == NULL)
		return NULL;
	memset(curnode, 0, sizeof(t_redir_node));
	curnode->mac = strdup(mac);

	if (prevnode == NULL) {
		first_redir_node = curnode;
	} else {
		prevnode->next = curnode;
	}
	
	return curnode;
}


void
free_redir_node(t_redir_node *node)
{

	if (node->mac != NULL)
		free(node->mac);

	free(node);
}


void
redir_list_delete(t_redir_node *node)
{
	t_redir_node *ptr;

	ptr = first_redir_node;

	if (ptr == NULL) {
		debug(LOG_ERR, "Node list empty!");
	} else if (ptr == node) {
		first_redir_node = ptr->next;
		free_redir_node(node);
	} else {
		/* Loop forward until we reach our point in the list. */
		while (ptr->next != NULL && ptr->next != node) {
			ptr = ptr->next;
		}
		/* If we reach the end before finding out element, complain. */
		if (ptr->next == NULL) {
			debug(LOG_ERR, "Node to delete could not be found.");
		/* Free element. */
		} else {
			ptr->next = node->next;
			free_redir_node(node);
		}
	}
}


void notify_client_connect(char *mac)
{
	t_client *client;
	t_redir_node *node;

	LOCK_REDIR();

	node = redir_list_find(mac);
	if (!node) {
		node = redir_list_append(mac);
	}
	if (!node) {
		UNLOCK_REDIR();
		return;
	}

	node->expiry = time(NULL);

	if (!node->redir_pending) {
		node->redir_pending = 1;
		if (redir_pending < 0) {
			debug(LOG_NOTICE, "Error: redir_pending = %d\n", redir_pending);
			redir_pending = 0;
		}
		redir_pending++;
		if (redir_pending == 1) {
			execute("echo 1 > /proc/sys/net/bridge/bridge-nf-enabled", 0);
		}
	}

	UNLOCK_REDIR();

	LOCK_CLIENT_LIST();

	client = client_list_find_by_mac(mac);
	if (client) {
		fw_deny(client->ip, client->mac, client->fw_connection_state);
		client_list_delete(client);
	}

	UNLOCK_CLIENT_LIST();
}

void notify_client_disconnect(char *mac)
{
	t_client *client;
	t_redir_node *node;

	LOCK_REDIR();

	node = redir_list_find(mac);
	if (node) {
		if (node->redir_pending) {
			if(redir_pending > 0) {
				redir_pending--;
			}
			if (redir_pending == 0) {
				execute("echo 0 > /proc/sys/net/bridge/bridge-nf-enabled", 0);
			}
			if (redir_pending < 0) {
				debug(LOG_NOTICE, "Error: redir_pending = %d\n", redir_pending);
				redir_pending = 0;
			}
		}
		redir_list_delete(node);
	}

	UNLOCK_REDIR();

	LOCK_CLIENT_LIST();
	client = client_list_find_by_mac(mac);
	if (client) {
		fw_deny(client->ip, client->mac, client->fw_connection_state);
		client_list_delete(client);
	}
	UNLOCK_CLIENT_LIST();
}

void
http_callback_404(httpd *webserver, request *r)
{
	char		*newlocation,
			*protocol,
			tmp_url[MAX_BUF],
			*url;
	int		port;
	s_config	*config = config_get_config();
	t_auth_serv	*auth_server = get_auth_server();

	debug(LOG_DEBUG, "----->http_callback_404");

	memset(tmp_url, 0, sizeof(tmp_url));
	snprintf(tmp_url, (sizeof(tmp_url) - 1), "http://%s%s",
			r->request.host,
			r->request.path);
	url = httpdUrlEncode(tmp_url);

	if (config->portal)
	{
		/* Re-direct them to auth server */
		safe_asprintf(&newlocation, "Location: %s", config->portal);
		httpdSetResponse(r, "302 Moved Temporarily\n");
		httpdAddHeader(r, newlocation);
		httpdPrintf(r, "\n");
		debug(LOG_INFO, "Captured %s requesting [%s] and re-directed them to portal page : %s", r->clientAddr, IS_NULL(url), config->portal);
		debug(LOG_INFO, "request host = %s, user agent = %s", IS_NULL(r->request.host), IS_NULL(r->request.userAgent));

		if ((strncasecmp(r->request.userAgent, "Mozilla",    7) == 0) ||
			(strncasecmp(r->request.userAgent, "Firefox",    7) == 0) ||
			(strncasecmp(r->request.userAgent, "MSIE",       4) == 0) ||
			(strncasecmp(r->request.userAgent, "Netscape",   8) == 0) ||
			(strncasecmp(r->request.userAgent, "Safari",     6) == 0) ||
			(strncasecmp(r->request.userAgent, "Opera",      5) == 0) ||
			(strncasecmp(r->request.userAgent, "Multizilla", 10) == 0)
		   )
		{
			http_allow_client(webserver, r);

			LOCK_REDIR();

			{
				t_redir_node *node;
				char *mac;

				if((mac = arp_get(r->clientAddr)) && 
					(node = redir_list_find(mac))) {
					if (node->redir_pending) {
						if(redir_pending > 0) {
							redir_pending--;
						}
						if (redir_pending == 0) {
							execute("echo 0 > /proc/sys/net/bridge/bridge-nf-enabled", 0);
						}
						if (redir_pending < 0) {
							debug(LOG_NOTICE, "Error: redir_pending = %d\n", redir_pending);
							redir_pending = 0;
						}
					}
					redir_list_delete(node);
				}
			}

			UNLOCK_REDIR();
		}
		free(newlocation);

	} else if (!is_online()) {
		/* The internet connection is down at the moment  - apologize and do not redirect anywhere */
		http_wifidog_header(r, "Uh oh! Internet access unavailable");
		httpdOutput(r, "We apologize, but it seems that the internet connection that powers this hotspot is temporarily unavailable.");
		httpdOutput(r, "<p>");
		httpdOutput(r, "If at all possible, please notify the owners of this hotspot that the internet connection is out of service.");
		httpdOutput(r, "<p>");
		httpdOutput(r, "The maintainers of this network are aware of this disruption.  We hope that this situation will be resolved soon.");
		httpdOutput(r, "<p>");
		httpdPrintf(r, "In a while please <a href='%s'>click here</a> to try your request again.", tmp_url);
		http_wifidog_footer(r);
		debug(LOG_INFO, "Sent %s an apology since I am not online - no point sending them to auth server", r->clientAddr);
	}
	else if (!is_auth_online()) {
		/* The auth server is down at the moment - apologize and do not redirect anywhere */
		http_wifidog_header(r, "Uh oh! Login screen unavailable");
		httpdOutput(r, "We apologize, but it seems that we are currently unable to re-direct you to the login screen.");
		httpdOutput(r, "<p>");
		httpdOutput(r, "The maintainers of this network are aware of this disruption.  We hope that this situation will be resolved soon.");
		httpdOutput(r, "<p>");
		httpdPrintf(r, "In a couple of minutes please <a href='%s'>click here</a> to try your request again.", tmp_url);
		http_wifidog_footer(r);
		debug(LOG_INFO, "Sent %s an apology since auth server not online - no point sending them to auth server", r->clientAddr);
	}
	else {
		if (auth_server->authserv_use_ssl) {
			protocol = "https";
			port = auth_server->authserv_ssl_port;
		} else {
			protocol = "http";
			port = auth_server->authserv_http_port;
		}

		/* Re-direct them to auth server */
		safe_asprintf(&newlocation, "Location: %s://%s:%d%slogin?gw_address=%s&gw_port=%d&gw_id=%s&url=%s",
			protocol,
			auth_server->authserv_hostname,
			port,
			auth_server->authserv_path,
			config->gw_address,
			config->gw_port, 
			config->gw_id,
			url);
		httpdSetResponse(r, "307 Please authenticate yourself here\n");
		httpdAddHeader(r, newlocation);
		http_wifidog_header(r, "Redirection");
		httpdPrintf(r, "Please <a href='%s://%s:%d%slogin?gw_address=%s&gw_port=%d&gw_id=%s&url=%s'>click here</a> to login",
				protocol,
				auth_server->authserv_hostname,
				port,
				auth_server->authserv_path,
				config->gw_address, 
				config->gw_port,
				config->gw_id,
				url);
		http_wifidog_footer(r);
		debug(LOG_INFO, "Captured %s requesting [%s] and re-directed them to login page", r->clientAddr, url);
		free(newlocation);
	}

	free(url);
}

void 
http_callback_wifidog(httpd *webserver, request *r)
{
	http_wifidog_header(r, "WiFiDog");
	httpdOutput(r, "Please use the menu on the left to navigate the features of this WiFiDog installation.");
	http_wifidog_footer(r);
}

void 
http_callback_about(httpd *webserver, request *r)
{
	http_wifidog_header(r, "About WiFiDog");
	httpdOutput(r, "This is WiFiDog version <b>" VERSION "</b>");
	http_wifidog_footer(r);
}

void 
http_callback_status(httpd *webserver, request *r)
{
	char * status = NULL;
	status = get_status_text();
	http_wifidog_header(r, "WiFiDog Status");
	httpdOutput(r, "<pre>");
	httpdOutput(r, status);
	httpdOutput(r, "</pre>");
	http_wifidog_footer(r);
	free(status);
}

void 
http_callback_auth(httpd *webserver, request *r)
{
	t_client	*client;
	httpVar * token;
	char	*mac;

	if ((token = httpdGetVariableByName(r, "token"))) {
		/* They supplied variable "token" */
		if (!(mac = arp_get(r->clientAddr))) {
			/* We could not get their MAC address */
			debug(LOG_ERR, "Failed to retrieve MAC address for ip %s", r->clientAddr);
			http_wifidog_header(r, "WiFiDog Error");
			httpdOutput(r, "Failed to retrieve your MAC address");
			http_wifidog_footer(r);
		} else {
			/* We have their MAC address */

			LOCK_CLIENT_LIST();
			
			if ((client = client_list_find(r->clientAddr, mac)) == NULL) {
				debug(LOG_DEBUG, "New client for %s", r->clientAddr);
				client_list_append(r->clientAddr, mac, token->value);
			} else {
				debug(LOG_DEBUG, "Node for %s already exists", client->ip);
			}

			UNLOCK_CLIENT_LIST();

			authenticate_client(r);
			free(mac);
		}
	} else {
		/* They did not supply variable "token" */
		http_wifidog_header(r, "WiFiDog Error");
		httpdOutput(r, "Invalid token");
		http_wifidog_footer(r);
	}
}

void
http_allow_client(httpd *webserver, request *r)

{
	t_client	*client;
	httpVar * token;
	char	*mac;

	if (!(mac = arp_get(r->clientAddr))) {
		/* We could not get their MAC address */
			debug(LOG_ERR, "Failed to retrieve MAC address for ip %s", r->clientAddr);
			http_wifidog_header(r, "WiFiDog Error");
			httpdOutput(r, "Failed to retrieve your MAC address");
			http_wifidog_footer(r);
	} else {
		/* We have their MAC address */
		LOCK_CLIENT_LIST();
		
		if ((client = client_list_find(r->clientAddr, mac)) == NULL) {
			debug(LOG_DEBUG, "New client for %s", r->clientAddr);
			client_list_append(r->clientAddr, mac, "");
		} else {
			debug(LOG_DEBUG, "Node for %s already exists", client->ip);
		}

		UNLOCK_CLIENT_LIST();

		permit_client(r);
		free(mac);
	}
}

void
http_wifidog_header(request *r, char *title)
{
	httpdOutput(r, "<html>\n");
	httpdOutput(r, "<head>\n");
	httpdPrintf(r, "<title>%s</title>\n", title);
	httpdOutput(r, "<meta HTTP-EQUIV='Pragma' CONTENT='no-cache'>\n");
	httpdOutput(r, "</head>\n");
	httpdOutput(r, "<body topmargin=0 leftmargin=0 marginwidth=0 marginheight=0 bgcolor=white text=#628C53 link=blue alink=blue vlink=blue>\n");

	httpdOutput(r, "<table width=100%% height=100%% border=0 cellpadding=12 cellspacing=5>\n");

	httpdOutput(r, "<tr>\n");

	httpdOutput(r, "<td valign=top align=right width=30%% bgcolor=#e1f5da>\n");
	httpdOutput(r, "&nbsp;<p>\n");
	httpdOutput(r, "&nbsp;<p>\n");
	httpdOutput(r, "<a href='/wifidog/status'>WiFiDog Status</a>\n");
	httpdOutput(r, "<p>\n");
	httpdOutput(r, "<a href='/wifidog/about'>About WiFiDog</a>\n");
	httpdOutput(r, "<p>\n");
	httpdOutput(r, "<a href='http://www.ilesansfil.org/wiki/WiFiDog'>WiFiDog's homepage</a>\n");
	httpdOutput(r, "</td>\n");

	httpdOutput(r, "<td valign=top align=left>\n");
	httpdPrintf(r, "<h1>%s</h1>\n", title);
	httpdOutput(r, "<hr>\n");

}

void
http_wifidog_footer(request *r)
{
	httpdOutput(r, "</td>\n");

	httpdOutput(r, "</tr>\n");

	httpdOutput(r, "<tr>\n");

	httpdOutput(r, "<td colspan=2 height=1 valign=bottom align=center>\n");
	httpdOutput(r, "<hr>\n");
	httpdOutput(r, "<font size=1>\n");
	httpdOutput(r, "Copyright (C) 2004-2005.  This software is released under the GNU GPL license.\n");
	httpdOutput(r, "</font>\n");
	httpdOutput(r, "</td>\n");

	httpdOutput(r, "</tr>\n");

	httpdOutput(r, "</table>\n");

	httpdOutput(r, "</body>\n");
	httpdOutput(r, "</html>\n");
}
