/*
 * br_hook.c
 *
 * Copyright (c) Arada Systems, 1996-2007
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/netfilter_bridge.h>
#include <linux/if_ether.h>
#include <linux/if_vlan.h>
#include <net/sock.h>
#include <linux/br_hook_defs.h>
#include "br_private.h"
#include "br_hook.h"

/*************** Macros ***************/
#define BR_RULE_NULL \
{ \
	.br = NULL, \
	.indev = NULL, \
	.outdev = NULL, \
	.pkt_type = 0, \
	.target = 0, \
	.arg = 0, \
}

#define BR_RULE_CLEAR(__br_rule) \
{ \
	__br_rule->br = NULL; \
	__br_rule->indev = NULL; \
	__br_rule->outdev = NULL; \
	__br_rule->pkt_type = 0; \
	__br_rule->target = 0; \
	__br_rule->arg = 0; \
}

#define BR_RULE_COPY(__dst, __src) \
{ \
	__dst->br = __src->br; \
	__dst->indev = __src->indev; \
	__dst->outdev = __src->outdev; \
	__dst->pkt_type = __src->pkt_type; \
	__dst->target = __src->target; \
	__dst->arg = __src->arg; \
}

#define BR_RULE_CMP(__rule, __match) \
({ \
	int __no_match = 0; \
	if (__match->br && __match->br != __rule->br) \
		__no_match = 1; \
	if (__match->indev && __match->indev != __rule->indev) \
		__no_match = 1; \
	if (__match->outdev && __match->outdev != __rule->outdev) \
		__no_match = 1; \
	if (__match->pkt_type && __match->pkt_type != __rule->pkt_type) \
		__no_match = 1; \
	if (__match->target && __match->target != __rule->target) \
		__no_match = 1; \
	if (__match->arg && __match->arg != __rule->arg) \
		__no_match = 1; \
 	__no_match; \
})

#define BR_RULE_PUSH(__br_hook) \
{ \
	int __i; \
	for (__i = __br_hook->num_rules; __i > 0; __i--) { \
		BR_RULE_COPY((&__br_hook->br_rules[__i]), (&__br_hook->br_rules[__i-1])); \
	} \
}

#define BR_RULE_PULL(__br_hook, __position) \
{ \
	int __i; \
	for (__i = __position; __i < __br_hook->num_rules-1; __i++) { \
		BR_RULE_COPY((&__br_hook->br_rules[__i]), (&__br_hook->br_rules[__i+1])); \
	} \
}


/*************** Forward Definitions ***************/
int add_8021q_tag_and_continue(struct sk_buff **pskb, int hook, int arg);
int rem_8021q_tag_and_continue(struct sk_buff **pskb, int hook, int arg);
int accept_eapol(struct sk_buff **pskb, int hook, int arg);


/*************** Globals ***************/
int (*br_targets[])(struct sk_buff **pskb, int hook, int arg) = 
{
	NULL,
	add_8021q_tag_and_continue, 
	rem_8021q_tag_and_continue,
	accept_eapol
};

struct bridge_hook br_hooks[] = 
{
	{
		.name = "BR_PREROUTING",
		.lock = SPIN_LOCK_UNLOCKED,
		.num_rules = 0,
		.br_rules = {BR_RULE_NULL,BR_RULE_NULL,BR_RULE_NULL,BR_RULE_NULL,
				BR_RULE_NULL,BR_RULE_NULL,BR_RULE_NULL,BR_RULE_NULL}
	},
	{
		.name = "BR_POSTROUTING",
		.lock = SPIN_LOCK_UNLOCKED,
		.num_rules = 0,
		.br_rules = {BR_RULE_NULL,BR_RULE_NULL,BR_RULE_NULL,BR_RULE_NULL,
				BR_RULE_NULL,BR_RULE_NULL,BR_RULE_NULL,BR_RULE_NULL}
	},
	{
		.name = "BR_INPUT",
		.lock = SPIN_LOCK_UNLOCKED,
		.num_rules = 0,
		.br_rules = {BR_RULE_NULL,BR_RULE_NULL,BR_RULE_NULL,BR_RULE_NULL,
				BR_RULE_NULL,BR_RULE_NULL,BR_RULE_NULL,BR_RULE_NULL}
	}
};


/*************** Utility functions ***************/
static int BR_HOOK_ADD(int hook, 
			int position,
			struct bridge_rule *br_rule)
{
	struct bridge_hook *br_hook = &br_hooks[hook];
	int i;

	spin_lock(&br_hook->lock);

	if (br_hook->num_rules >= MAX_RULES) {
		spin_unlock(&br_hook->lock);
		printk(KERN_ERR "br_hook %s: No place left to add new rule\n", br_hook->name);
		return -1;
	}

	for (i = 0; i < br_hook->num_rules; i++) {
		if (!BR_RULE_CMP((&br_hook->br_rules[i]), br_rule)) {
			spin_unlock(&br_hook->lock);
			printk(KERN_ERR "br_hook %s: Rule already exists\n", br_hook->name);
			return -1;
		}
	}

	switch (position) {
		case BR_FIRST:
			BR_RULE_PUSH(br_hook);
			BR_RULE_COPY((&br_hook->br_rules[0]), br_rule);
			br_hook->num_rules++;
			break;
		case BR_NEXT:
			BR_RULE_COPY((&br_hook->br_rules[br_hook->num_rules]), br_rule);
			br_hook->num_rules++;
			break;
		default:
			printk(KERN_ERR "br_hook: Unknown position %d to add a rule\n", position);
			spin_unlock(&br_hook->lock);
			return -1;
	}	

	spin_unlock(&br_hook->lock);

	return 0;
}

static int BR_HOOK_DEL(int hook, 
			struct bridge_rule *br_rule)
{
	struct bridge_hook *br_hook = &br_hooks[hook];
	int i;

	spin_lock(&br_hook->lock);

	if (br_hook->num_rules == 0) {
		spin_unlock(&br_hook->lock);
		printk(KERN_ERR "br_hook %s: No more rules to delete\n", br_hook->name);
		return -1;
	}

	while (br_hook->num_rules) {
		int deleted = 0;
		for (i = 0; i < br_hook->num_rules; i++) {
			if (!BR_RULE_CMP((&br_hook->br_rules[i]), br_rule)) {
				BR_RULE_PULL(br_hook, i);
				br_hook->num_rules--;
				BR_RULE_CLEAR((&br_hook->br_rules[br_hook->num_rules]));
				deleted = 1;
				break;
			}
		}
		if (!deleted)
			break;
	}

	spin_unlock(&br_hook->lock);

	return 0;
}

int add_br_hook_rule(int hook,
			int position,
			char *bridge, 
			char *indev, 
			char *outdev, 
			unsigned int pkt_type, 
			int target,
			int arg)
{
	struct net_bridge *br = NULL;
	struct net_device *br_dev = NULL;
	struct net_device *in = NULL;
	struct net_device *out = NULL;
	struct bridge_rule br_rule;

	if (strcmp(bridge, "*")) {
		br_dev = dev_get_by_name(bridge);
		if (!br_dev)
			return -1;
		br = (struct net_bridge *)netdev_priv(br_dev);
		dev_put(br_dev);
	}

	if (strcmp(indev, "*")) {
		in = dev_get_by_name(indev);
		if (!in)
			return -1;
		dev_put(in);
	}

	if (strcmp(outdev, "*")) {
		out = dev_get_by_name(outdev);
		if (!out)
			return -1;
		dev_put(out);
	}

	br_rule.br = br;
	br_rule.indev = in;
	br_rule.outdev = out;
	br_rule.pkt_type = pkt_type;
	br_rule.target = target;
	br_rule.arg = arg;

	return BR_HOOK_ADD(hook, position, &br_rule);
}

int del_br_hook_rule(int hook,
			char *bridge,
			char *indev,
			char *outdev,
			unsigned int pkt_type,
			int target,
			int arg)
{
	struct net_bridge *br = NULL;
	struct net_device *br_dev = NULL;
	struct net_device *in = NULL;
	struct net_device *out = NULL;
	struct bridge_rule br_rule;

	if (strcmp(bridge, "*")) {
		br_dev = dev_get_by_name(bridge);
		if (!br_dev)
			return -1;
		br = (struct net_bridge *)netdev_priv(br_dev);
		dev_put(br_dev);
	}

	if (strcmp(indev, "*")) {
		in = dev_get_by_name(indev);
		if (!in)
			return -1;
		dev_put(in);
	}

	if (strcmp(outdev, "*")) {
		out = dev_get_by_name(outdev);
		if (!out)
			return -1;
		dev_put(out);
	}

	br_rule.br = br;
	br_rule.indev = in;
	br_rule.outdev = out;
	br_rule.pkt_type = pkt_type;
	br_rule.target = target;
	br_rule.arg = arg;

	return BR_HOOK_DEL(hook, &br_rule);
}

int get_next_br_hook_rule(int hook,
				int *cookie,
				char *bridge,
				char *indev,
				char *outdev,
				unsigned int *pkt_type,
				int *target,
				int *arg)
{
	struct bridge_hook *br_hook = &br_hooks[hook];
	struct bridge_rule *br_rule;

	spin_lock(&br_hook->lock);
	
	if (*cookie >= 0 && *cookie < br_hook->num_rules) {
		br_rule = &br_hook->br_rules[*cookie];
		(*cookie)++;

		if (br_rule->br)
			strcpy(bridge, br_rule->br->dev->name);
		else
			strcpy(bridge, "*");

		if (br_rule->indev)
			strcpy(indev, br_rule->indev->name);
		else
			strcpy(indev, "*");

		if (br_rule->outdev)
			strcpy(outdev, br_rule->outdev->name);
		else
			strcpy(outdev, "*");

		*pkt_type = br_rule->pkt_type;

		*target = br_rule->target;

		*arg = br_rule->arg;
	} else {
		*cookie = -1;
	}

	spin_unlock(&br_hook->lock);

	return 0;
}


/*************** Taget functions ***************/
/* Add 802_1Q tag */
int add_8021q_tag_and_continue(struct sk_buff **pskb,
				int hook,
				int arg)
{
	struct vlan_ethhdr *veth;

	if (hook != BR_PREROUTING)
		return BR_CONTINUE;
	
	/* VLAN_ETH_HLEN = ETH_HLEN (6 + 6 + 2) + VLAN_HLEN (4) */
        if (skb_headroom(*pskb) < VLAN_ETH_HLEN) {
		struct sk_buff *sk_tmp = *pskb;

		//printk("%s#%d\n", __FUNCTION__, __LINE__);
		*pskb = skb_realloc_headroom(sk_tmp, VLAN_ETH_HLEN);
		kfree_skb(sk_tmp);
		if (!*pskb) {
			printk(KERN_ERR "%s#%d: failed to realloc headroom\n",
					__FILE__, __LINE__);
			return BR_DROP;
		}
        } else {
		if (skb_shared((*pskb)) || skb_cloned((*pskb))) {
			struct sk_buff *nskb;

			//printk("%s#%d\n", __FUNCTION__, __LINE__);
			nskb = skb_copy((*pskb), GFP_ATOMIC);
			if (!nskb) {
				printk(KERN_ERR "%s#%d: failed to copy skb\n",
					__FILE__, __LINE__);
				kfree_skb(*pskb);
				return BR_DROP;
			}
			if ((*pskb)->sk)
				skb_set_owner_w(nskb, (*pskb)->sk);
			kfree_skb((*pskb));
			*pskb = nskb;
		}
	}

	/* Add 4 bytes for VLAN header */
	veth = (struct vlan_ethhdr *)skb_push(*pskb, VLAN_ETH_HLEN);
	/* Move the mac addresses to the beginning of the new header. */
	memmove((*pskb)->data, (*pskb)->data + VLAN_HLEN, (2 * ETH_ALEN));

	/* first, the ethernet type */
	veth->h_vlan_proto = __constant_htons(ETH_P_8021Q);
	/* now, the tag */
	veth->h_vlan_TCI = htons((unsigned short)arg);
	(*pskb)->protocol = __constant_htons(ETH_P_8021Q);
	(*pskb)->mac.raw = (*pskb)->data;
	/* Redo the side effects of eth_type_trans, called by lower drivers */
	skb_pull(*pskb, ETH_HLEN);
	(*pskb)->h.raw = (*pskb)->nh.raw = (*pskb)->data;

	return BR_CONTINUE;
}

/* Remove 802_1Q tag */
int rem_8021q_tag_and_continue(struct sk_buff **pskb, 
				int hook, 
				int arg)
{
	struct vlan_ethhdr *veth;

	if (hook != BR_POSTROUTING)
		return BR_CONTINUE;
	
	veth = vlan_eth_hdr(*pskb);
	if ((ntohs(veth->h_vlan_TCI) & VLAN_VID_MASK) != (unsigned short)arg)
		return BR_CONTINUE;

	if (skb_shared(*pskb) || skb_cloned(*pskb)) {
		struct sk_buff *nskb;

		//printk("%s#%d\n", __FUNCTION__, __LINE__);
		nskb = skb_copy(*pskb, GFP_ATOMIC);
		if (!nskb) {
			kfree_skb(*pskb);
			return NF_DROP;
		}
		if ((*pskb)->sk)
			skb_set_owner_w(nskb, (*pskb)->sk);
		kfree_skb(*pskb);
		*pskb = nskb;
	}

	/* Gobble up the VLAN header by moving the ethernet src & dst addresses
	 * to right by 4 bytes
	 */
	memmove(eth_hdr(*pskb)->h_dest + VLAN_HLEN, eth_hdr(*pskb)->h_dest, (2 * ETH_ALEN));
	/* Eat 4 bytes at the head */
	skb_pull(*pskb, VLAN_HLEN);
	(*pskb)->mac.raw = (*pskb)->data;
	(*pskb)->protocol = eth_hdr(*pskb)->h_proto;

	return BR_CONTINUE;
}

/* Accept EAPOL frames */
int accept_eapol(struct sk_buff **pskb, 
		int hook, 
		int arg)
{
	if (hook != BR_PREROUTING)
		return BR_CONTINUE;

	return BR_ACCEPT;
}
