/*
 *  ebt_add1q
 *
 *  Copyright (c) Arada Systems, 1996-2007
 *
 */

#include <linux/module.h>
#include <linux/if_ether.h>
#include <linux/if_vlan.h>
#include <net/sock.h>
#include <linux/netfilter_bridge/ebtables.h>
#include "ebt_8021q.h"

static int ebt_target_add1q(struct sk_buff **pskb, unsigned int hooknr,
   const struct net_device *in, const struct net_device *out,
   const void *data, unsigned int datalen)
{
	struct ebt_8021q_info *info = (struct ebt_8021q_info *) data;
	struct vlan_ethhdr *veth;

	/* VLAN_ETH_HLEN = ETH_HLEN (6 + 6 + 2) + VLAN_HLEN (4) */

        if (skb_headroom(*pskb) < VLAN_ETH_HLEN) {
		struct sk_buff *sk_tmp = *pskb;

		*pskb = skb_realloc_headroom(sk_tmp, VLAN_ETH_HLEN);
		kfree_skb(sk_tmp);
		if (!*pskb) {
			printk(KERN_ERR "%s#%d: failed to realloc headroom\n",
					__FILE__, __LINE__);
			return NF_DROP;
		}
        } else {
		if (skb_shared((*pskb)) || skb_cloned((*pskb))) {
			struct sk_buff *nskb;

			nskb = skb_copy((*pskb), GFP_ATOMIC);
			if (!nskb) {
				printk(KERN_ERR "%s#%d: failed to copy skb\n",
					__FILE__, __LINE__);
				return NF_DROP;
			}
			if ((*pskb)->sk)
				skb_set_owner_w(nskb, (*pskb)->sk);
			kfree_skb((*pskb));
			*pskb = nskb;
		}
	}

	/* Add 4 bytes for VLAN header */
	veth = (struct vlan_ethhdr *)skb_push((*pskb), VLAN_ETH_HLEN);
	/* Move the mac addresses to the beginning of the new header. */
	memmove((*pskb)->data, (*pskb)->data + VLAN_HLEN, (2 * ETH_ALEN));

	/* first, the ethernet type */
	veth->h_vlan_proto = __constant_htons(ETH_P_8021Q);
	/* now, the tag */
	veth->h_vlan_TCI = htons(info->vlan_id);
	(*pskb)->protocol = __constant_htons(ETH_P_8021Q);

	(*pskb)->mac.raw = (*pskb)->data;
	/* Redo the side effects of eth_type_trans, called by lower drivers */
	skb_pull((*pskb), ETH_HLEN);
	(*pskb)->h.raw = (*pskb)->nh.raw = (*pskb)->data;

	return info->target;
}

static int ebt_target_add1q_check(const char *tablename, unsigned int hookmask,
   const struct ebt_entry *e, void *data, unsigned int datalen)
{
	struct ebt_8021q_info *info = (struct ebt_8021q_info *) data;

	if (datalen != EBT_ALIGN(sizeof(struct ebt_8021q_info)))
		return -EINVAL;
	if (BASE_CHAIN && info->target == EBT_RETURN)
		return -EINVAL;
	if(info->vlan_id > 4094)
		return -EINVAL;
	CLEAR_BASE_CHAIN_BIT;
	if (strcmp(tablename, "nat"))
		return -EINVAL;
	if (hookmask & ~(1 << NF_BR_PRE_ROUTING))
		return -EINVAL;
	if (INVALID_TARGET)
		return -EINVAL;
	return 0;
}

static struct ebt_target add1q =
{
	.name		= EBT_ADD1Q_TARGET,
	.target		= ebt_target_add1q,
	.check		= ebt_target_add1q_check,
	.me		= THIS_MODULE,
};

static int __init init(void)
{
	return ebt_register_target(&add1q);
}

static void __exit fini(void)
{
	ebt_unregister_target(&add1q);
}

module_init(init);
module_exit(fini);
MODULE_LICENSE("GPL");
